#!env python
from argparse import ArgumentParser, Namespace
import os
import torch as t
import json
from pytorch_lightning import Trainer
from pytorch_lightning.callbacks import EarlyStopping
from .plot import plot
from .tune_hyperparameters import tune

# FIXME: cannot run snapshost due to name


def run():
    current_dir = os.path.dirname(os.path.realpath(__file__))
    parser = ArgumentParser()
    cur_folder = os.path.dirname(__file__)
    current_path = os.getcwd()
    root_folder = os.path.basename(current_path)
    os.chdir("..")
    parser.add_argument(
        "action",
        help="train or test the model",
        choices=(
            "init",
            "train",
            "test",
            "list",
            "remove",
            "snapshot-create",
            "snapshot-run",
            "snapshot-list",
            "snapshot-remove",
            "tune",
            "create",
            "restart",
            "clone",
            "merge-results",
            "exec",
        ),
    )
    parser.add_argument(
        "model",
        default="",
        type=str,
        nargs="?",
        help=f"Model to use. Must exist in {root_folder}/models/",
    )
    parser.add_argument(
        "target",
        help="target name to use. Used by 'clone' and 'snapshot'",
        nargs="?",
        default=os.path.join(current_dir, "results"),
    )
    args = parser.parse_args()
    models_choices = (
        tuple(
            x
            for x in os.listdir(os.path.join(root_folder, "models"))
            if not "__" in x
        )
        if os.path.exists(os.path.join(root_folder, "models"))
        else tuple()
    )

    # assert "__" not in params.model, "Models cannot have '__' in their name"

    cur_folder = os.path.basename(os.path.dirname(__file__))

    if args.action == "init":
        os.system("touch .mateconfig.json")

    elif args.action == "create":
        os.mkdir(f"{os.path.join(root_folder, 'models', args.model)}")
        target_dir = os.path.join(root_folder, "models", args.model)
        os.system(
            f"cp -r {os.path.join(cur_folder, root_folder, 'example')}* {os.path.join(root_folder, 'models', args.model)}"
        )
        print(f"Created model {args.model} at {target_dir}")

    elif args.action == "remove":
        action = "go"
        while action not in ("y", "n"):
            action = input(
                f'Are you sure you want to remove model "{args.model}"? (y/n)\n'
            )
        if action == "y":
            os.system(
                f"rm -r {os.path.join(root_folder, 'models', args.model)}"
            )
            print(f"Removed model {args.model}")
        else:
            print("Ok, exiting.")

    elif args.action == "clone":
        os.system(
            f"cp -r {os.path.join(root_folder, 'models', args.model)} {os.path.join(root_folder, 'models', args.target)}"
        )

    elif args.action == "list":
        print("\n".join(f"\t{u}" for u in models_choices))

    elif args.action == "snapshot-create":
        snapshot_names = [
            name.split("__")
            for name in os.listdir(os.path.join(root_folder, "snapshots"))
        ]
        matching_snapshots = [
            name for name in snapshot_names if name[0] == args.model
        ]
        max_version_matching = (
            max([int(name[1]) for name in matching_snapshots])
            if len(matching_snapshots) > 0
            else 0
        )
        snapshot_name = f"{args.model}__{max_version_matching + 1}"
        os.system(
            f"cp -r {os.path.join(root_folder, 'models', args.model)} {os.path.join(root_folder, 'snapshots', snapshot_name)}"
        )
        print(f"Created snapshot {snapshot_name}")

    elif args.action == "snapshot-list":
        print(os.listdir(os.path.join(root_folder, "snapshots")))

    else:
        assert (
            args.model in models_choices
        ), f"Model {args.model} does not exist."
        with open(
            os.path.join(
                root_folder, "models", args.model, "default_parameters.json"
            )
        ) as f:
            params = Namespace()
            params.__dict__ = json.load(f)

        # for key, val in model_specific_params.items():
        #    # if key not in params.__dict__:
        #    args.__dict__[key] = val

        print(json.dumps(params.__dict__, indent=4))
        model_module = __import__(
            f"dl.{'models' if args.action != 'snapshot-run' else 'snapshots'}.{args.model}.model",
            fromlist=["models"],
        )
        data_loader_module = __import__(
            f"{root_folder}.data_loader", fromlist=["data_loader"]
        )
        logger_module = __import__(f"{cur_folder}.logger", fromlist=["logger"])
        save_path = os.path.join(root_folder, "models", args.model)
        params.save_path = save_path

        if args.action in ("train", "test", "restart", "exec"):
            model = model_module.Model(params)
            print(model)
            if not args.action == "exec":
                data_module = data_loader_module.CustomDataModule(params)
                trainer = Trainer(
                    max_epochs=params.max_epochs,
                    gpus=(1 if params.cuda else None),
                    callbacks=[
                        EarlyStopping(
                            monitor="val_loss",
                            patience=params.early_stopping_patience,
                        ),
                        # checkpoint_callback,
                    ],
                    logger=logger_module.CustomLogger(params),
                    enable_checkpointing=False,
                )

        checkpoint_path = os.path.join(save_path, "checkpoint.ckpt")
        if args.action == "train" and os.path.exists(checkpoint_path):
            action = "go"
            while action not in ("y", "n", ""):
                action = input(
                    "Checkpiont file exists. Re-training will erase it. Continue? ([y]/n)\n"
                )
            if action in ("y", ""):
                os.remove(checkpoint_path)
            else:
                print("Ok, exiting.")
                return

        if args.action in ("restart", "test", "exec"):
            if os.path.exists(checkpoint_path):
                print(f"\nLoading model from checkpoint:'{checkpoint_path}'\n")
                model.load_state_dict(t.load(checkpoint_path))
            else:
                raise Exception(
                    "No checkpoint found. You must train the model first!"
                )
        if args.action == "tune":
            tune(
                params, model_module.Model, data_loader_module.CustomDataModule
            )

        if args.action in ("train", "restart"):
            trainer.fit(model=model, datamodule=data_module)
            print(f"Saving model to {checkpoint_path}")
            t.save(model.cpu().state_dict(), checkpoint_path)
        if args.action in ("test", "train", "restart"):
            trainer.test(model=model, datamodule=data_module)
            plot(save_path)
        elif args.action == "exec":
            assert (
                "exec" in params.__dict__
            ), 'No "exec" file defined in parameters.json file.'
            custom_exec = __import__(
                f"{root_folder}.{params.exec}", fromlist=["custom_test"],
            ).main
            custom_exec(model)

        os.chdir(current_path)


if __name__ == "__main__":
    run()



from pytorch_lightning.loggers import LightningLoggerBase
import torch as t
import json
import csv
import os
import ipdb
from argparse import Namespace


class CustomLogger(LightningLoggerBase):
    def __init__(
        self,
        params: Namespace,
        train_header: tuple[str, ...] = ("epoch", "train_mse", "val_mse"),
    ):
        super().__init__()
        self.params = params
        self.csv_path = os.path.join(params.save_path, "training_log.csv")
        self.last_train_mse_loss = -1.0
        with open(self.csv_path, "w") as f:
            writer = csv.writer(f)
            writer.writerow(train_header)

    def write_performance_csv(self, name, metrics):
        if name == "test":
            metrics['epoch'] = 1
        performance_name = f"{name}_performance"
        file_path = os.path.join(
            self.params.save_path, f"{performance_name}.csv"
        )
        if metrics['epoch'] > 0:
            if metrics['epoch'] == 1:
                if os.path.exists(file_path):
                    os.remove(file_path)
                with open(file_path, "w") as f:
                    writer = csv.writer(f)
                    writer.writerow(metrics[performance_name].keys())

            with open(file_path, "a") as f:
                writer = csv.writer(f)
                writer.writerow(tuple(val for _, val in metrics[performance_name].items()))

    def log_metrics(self, metrics, step: int):
        for condition in ["train", "val", "test"]:
            if f"{condition}_performance" in metrics:
                self.write_performance_csv(condition, metrics)

    @property
    def name(self):
        return "CustomLogger"

    def log_hyperparams(self, hparams):
        pass

    @property
    def version(self):
        pass



import pytorch_lightning as pl
import os
from argparse import Namespace
import argparse
import os
from typing import Optional
import optuna
from optuna.integration import PyTorchLightningPruningCallback
from packaging import version
import pytorch_lightning as pl
import torch
from copy import deepcopy
from optuna.visualization import plot_optimization_history


def tune(params: Namespace, model_class, data_loader_class):
    params.max_epochs = 1

    def objective(trial: optuna.trial.Trial) -> float:

        # We optimize the number of layers, hidden units in each layer and dropouts.
        train_batch_size = trial.suggest_int("train_batch_size", 2, 32)
        kernel_size = trial.suggest_int("kernel_size", 3, 7)
        local_params = deepcopy(params)
        local_params.train_batch_size = train_batch_size
        local_params.kernel_size = kernel_size
        model = model_class(local_params)
        datamodule = data_loader_class(local_params)

        trainer = pl.Trainer(
            logger=True,
            enable_checkpointing=False,
            max_epochs=params.max_epochs,
            gpus=1 if torch.cuda.is_available() else None,
            callbacks=[
                PyTorchLightningPruningCallback(trial, monitor="val_acc")
            ],
        )
        values = dict(train_batch_size=train_batch_size)
        hyperparameters = Namespace()
        hyperparameters.__dict__ = values
        trainer.logger.log_hyperparams(hyperparameters)
        trainer.fit(model, datamodule=datamodule)

        return trainer.callback_metrics["val_acc"].item()

    args = Namespace()
    args.pruning = True
    pruner: optuna.pruners.BasePruner = (
        optuna.pruners.MedianPruner()
        if args.pruning
        else optuna.pruners.NopPruner()
    )
    study = optuna.create_study(direction="maximize", pruner=pruner)
    study.optimize(objective, n_trials=100, timeout=600)

    print(f"Number of finished trials: {len(study.trials)}")

    print("Best trial:")
    trial = study.best_trial

    print("  Value: {}".format(trial.value))

    print("  Params: ")
    for key, value in trial.params.items():
        print("    {}: {}".format(key, value))

    plot_optimization_history(study)
    return objective



from setuptools import setup, find_packages

"""
setup(
    name="yerbamate",
    version="0.3.2",
    license="MIT",
    author="Giulio Zani, Ali Rahimi",
    author_email="giulio.zani@gmail.com",
    scripts=["src/run.py"],
    packages=["src/yerbamate"],
    url="https://github.com/giuliozani/mate",
    keywords="deep_learning machine_learning package_manager",
    install_requires=["torch", "pytorch_lightning", "numpy",],
)
"""
setup(
    name="ymdl",
    description="A cowsay clone for python in one file.",
    author="Giulio Zani, Ali Rahimi",
    author_email="yerba.mate.dl@proton.me",
    url="https://github.com/ilex-paraguariensis/yerbamate",
    python_requires=">=3.9",
    version="0.1.1",
    packages=["yerbamate"],
    license="Apache License 2.0",
    license_files=("LICENSE.md",),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
    ],
    scripts=["./src/mate"],
)



import os
import pandas as pd
import ipdb
import torch as t
import matplotlib.pyplot as plt


def plot(path: str):
    val_file = os.path.join(path, "val_performance.csv")
    train_file = os.path.join(path, "train_performance.csv")
    test_file = os.path.join(path, "test_performance.csv")
    val_df = pd.read_csv(val_file)
    keys = list(val_df.keys())
    val = t.from_numpy(val_df.values).T
    # test = t.from_numpy(pd.read_csv(test_file).values).squeeze()
    train = t.from_numpy(pd.read_csv(train_file).values).T
    xs = t.arange(val.shape[1])
    train = train[:, : val.shape[1]]
    for i in range(val.shape[0]):
        plt.plot(xs, train[i], "--", label=f"train_{keys[i]}")
        plt.plot(xs, val[i], "--", label=f"val_{keys[i]}")
    # plt.plot([val.shape[1]], [test], "o", label="test")
    plt.legend()
    plt.title("Training")
    plt.savefig(os.path.join(path, "plot.png"))
